<template>
	<div class="mb-8">
		<a
			href="https://github.com/sponsors/privatenumber/sponsorships?pay_prorated=false&tier_id=388346"
			target="_blank"
			rel="noopener"
			class="sponsor-button"
		>
			Become a premium sponsor
		</a>
	</div>
</template>

<style scoped>
.sponsor-button {
	@apply
		block
		w-fit
		mx-auto
		text-xs
		rounded-full
		transition-colors
		bg-pink-500
		hover:bg-pink-600
		text-white
		mt-5
		py-1.5
		px-4;
}
</style>
