@tailwind base;
@tailwind components;
@tailwind utilities;

.button {
	@apply
		inline-block
		py-2
		px-6
		rounded-full
		!no-underline
		!transition-colors
		duration-200
		;
}

.dark .dark-zebra-pattern {
	background-image: repeating-linear-gradient(
		45deg,
		transparent,
		transparent 10px,
		rgb(0 0 0/ 0.05) 10px,
		rgb(0 0 0/ 0.05) 20px
	);
}
