export {
	register,
	type InitializationOptions,
	type NamespacedUnregister,
	type Register,
	type RegisterOptions,
	type Unregister,
} from './register.js';
export type { ScopedImport } from './scoped-import.js';
export { tsImport } from './ts-import.js';
